@@ -2,7 +2,7 @@
 # Copyright 2012 by Kay-Uwe (Kiwi) Lorenz
 # Published under New BSD License, see LICENSE.txt for details.
 #
-from ctxvar import ctx, ctxget, Undef, Here
+from .util import Undef
 import re, sys, os, logging, time
 from .rules import *
 from .grammar_tester import TokenTest, RuleTest
@@ -57,7 +57,9 @@ def native_rule_producer(lines):
             indent = re.match("^\s*", line).group(0)
             continue
 
-        if re.match("^%(indent)s\s"%locals(), line):
+        if re.match("^\s*#", line): continue
+
+        if re.match("^%(indent)s\s(?!\s*#)"%locals(), line):
             rule += line
             continue
 
@@ -81,7 +83,7 @@ def sphinx_rule_producer(lines):
 
     do_grammar = True
 
-    highlight_aptk = True
+    highlight_aptk = False
 
     line_no = 0
     for line in lines:
@@ -135,12 +137,14 @@ def sphinx_rule_producer(lines):
             else: continue
 
         if rule is None:
-            rule = line
+            rule = line+"\n"
             indent = re.match("^\s*", line).group(0)
             continue
 
-        if re.match("^%(indent)s\s"%locals(), line):
-            rule += line
+        if re.match("^\s*#", line): continue
+
+        if re.match("^%(indent)s\s(?!\s*#)"%locals(), line):
+            rule += line+"\n"
             continue
 
         if not re.match("^%(indent)s"%locals(), line):
@@ -153,7 +157,7 @@ def sphinx_rule_producer(lines):
 
         elif rule is not None:
             yield (rule, line_no)
-            rule = line
+            rule = line+"\n"
             indent = re.match("^\s*", line).group(0)
 
     if rule is not None:
@@ -185,10 +189,10 @@ class GrammarCompiler:
     RE_ARG_RULE     = re.compile(r'<\.?\w+:(?:[^>\\]|\\\\|\\>)*>')
     RE_QUANT        = re.compile(r'\{\d*,\d*\}|\{\d+\}$')
 
-    RE_DQ_STRING      = re.compile(r'"((?:\\\\|\\"|[^"\\]+)*)"')
-    RE_SQ_STRING      = re.compile(r"'((?:\\\\|\\'|[^'\\]+)*)'")
-    RE_DQ_STRING_ONLY = re.compile(r'"(?:\\\\|\\"|[^"\\]+)*"$')
-    RE_SQ_STRING_ONLY = re.compile(r"'(?:\\\\|\\'|[^'\\]+)*'$")
+    RE_DQ_STRING      = re.compile(r'"((?:\\\\|\\[^\\]|[^"\\]+)*)"')
+    RE_SQ_STRING      = re.compile(r"'((?:\\\\|\\[^\\]|[^'\\]+)*)'")
+    RE_DQ_STRING_ONLY = re.compile(r'"((?:\\\\|\\[^\\]|[^"\\]+)*)"$')
+    RE_SQ_STRING_ONLY = re.compile(r"'((?:\\\\|\\[^\\]|[^'\\]+)*)'$")
 
     doc_string_filter = [ remove_first_line, remove_indent ]
     doc_string_type   = 'sphinx'
@@ -352,11 +356,12 @@ class GrammarCompiler:
             self.grammar._BACKTRACK_ = False
 
         action = None
-        if operator[0] == '(':
-            if operator[-2:] == ')=' :
-                action = 'make_'+operator[1:-2]
-        elif operator[:-1] in PARSE_ACTION_MAP:
-            action = PARSE_ACTION_MAP[operator[0]]
+
+        op = operator[:-1]
+        if op in PARSE_ACTION_MAP:
+            action = PARSE_ACTION_MAP[op]
+        elif op != ':':
+            action = op
 
         if action:
             if pair:
@@ -375,7 +380,9 @@ class GrammarCompiler:
             #old_sigspace = self.grammar._SIGSPACE_
             #self.grammar._SIGSPACE_ = True
 
-        setattr(G, name, self.alternatives(tokens))
+        setattr(G, name, self.alternatives(tokens, 
+            backtrack=self.grammar._BACKTRACK_, 
+            sigspace=self.grammar._SIGSPACE_))
 
         #if sigspace is not None:
             #self.grammar._SIGSPACE_ = old_sigspace
@@ -404,12 +411,14 @@ class GrammarCompiler:
         m = self.RE_DQ_STRING.match(s)
         if m:
            p = m.end()
-           input = re.sub(r'\\(\\|")', r'\1', m.group(1))
+           #input = re.sub(r'\\(\\|")', r'\1', m.group(1))
+           input = m.group(1).decode('string_escape')
         else:
            m = self.RE_SQ_STRING.match(s)
            if m:
                p = m.end()
-               input = re.sub(r"\\(\\|')", r'\1', m.group(1))
+               #input = re.sub(r"\\(\\|')", r'\1', m.group(1))
+               input = m.group(1).decode('string_escape')
            else:
                m = re.match(r'(?:\|[^\n]*\n)+', s)
                if m:
@@ -431,6 +440,7 @@ class GrammarCompiler:
         expected = None
 
         p = s.find('-', p)
+
         if p >= 0:
           if s.startswith('->', p):
             parse_actions = None
@@ -451,11 +461,14 @@ class GrammarCompiler:
 
           m = self.RE_DQ_STRING_ONLY.match(s.rstrip())
           if m:
-            expected = re.sub(r'\\(\\|")', r'\1', m.group(1))
+            log.error("m: %s", m.groups())
+            #expected = re.sub(r'\\(\\|")', r'\1', m.group(1))
+            expected = m.group(1).decode('string_escape')[1:-1]
           else:
             m = self.RE_SQ_STRING_ONLY.match(s.rstrip())
             if m:
-                expected = re.sub(r"\\(\\|')", r'\1', m.group(1))
+               # expected = re.sub(r"\\(\\|')", r'\1', m.group(1))
+                expected = m.group(1).decode('string_escape')[1:-1]
             else:
                 m = re.match(r'(?:\|[^\n]*\n)+', s)
                 if m:
@@ -463,16 +476,26 @@ class GrammarCompiler:
                 else:
                     expected = s.strip()
 
+
         if name[0] == '{' and name[-1] == '}':
-            self.grammar._TESTS_.append(TokenTest(name[1:-1], op, pos,
-                input, parse_actions, expected, skip=self.grammar._SKIP_))
+            test = TokenTest
+            name = name[1:-1]
 
         elif name[0] == '<' and name[-1] == '>':
-            self.grammar._TESTS_.append(RuleTest(name[1:-1], op, pos,
-                input, parse_actions, expected, skip=self.grammar._SKIP_))
+            test = RuleTest
+            name = name[1:-1]
+
         else:
-            self.grammar._TESTS_.append(RuleTest(name, op, pos,
-                input, parse_actions, expected, skip=self.grammar._SKIP_))
+            test = RuleTest
+
+        debug = self.grammar._TMP_.get('debugging', 0)
+        if self.grammar._TMP_.get('debug', 0):
+            debug = 1
+            del self.grammar._TMP_['debug']
+
+        self.grammar._TESTS_.append(test(name, 
+            op, pos, input, parse_actions, expected,
+            skip=self.grammar._SKIP_, debug=debug))
 
     def definition(self, s):
         '''handle token definition and rule definition'''
@@ -508,6 +531,8 @@ class GrammarCompiler:
                     if _name is None:
                         _name = t
                         if len(_name) > 2 and _name.startswith('"') and _name.endswith('"'): _name = _name[1:-1]
+                    elif t == '=>':
+                        continue
                     else:
                         self.grammar._PARSE_ACTION_MAP_[_name] = t
                         _name = None
@@ -521,20 +546,40 @@ class GrammarCompiler:
                     else:
                         pa = self.grammar._PARSE_ACTIONS_
                         x = {}
-                        assert re.match("^\w+(\.\w+)+$", t)
-                        mod, thing = t.rsplit('.', 1)
+
+                        m = re.match(r"^\.?(\w+)$", t)
+                        if m:
+                            mod = self.grammar.__module__
+                            thing = m.group(1)
+                        else:
+                            assert re.match(r"^\w+(\.\w+)+$", t), "%s is not valid" % t
+                            mod, thing = t.rsplit('.', 1)
+
                         exec "from %s import %s" % (mod, thing) in x
-                        if callable(x[thing]): pa[_name] = x[thing]()
-                        else: pa[_name] = x[thing]
+                        parse_action_factory = x[thing]
+
+                        if callable(parse_action_factory): pa[_name] = parse_action_factory()
+                        else: pa[_name] = parse_action_factory
 
             elif tokens[0] == ":args-of":
                 _name = tokens[1]
                 _list = tokens[2:]
+
+                postprocessor = None    # postprocessor may return a rule, 
+                if len(_list) > 2 and _list[-2] == '=>':
+                    postprocessor = _list[-1]
+                    _list = _list[:-2]
+
                 for e in _list:
                     assert e in ("string", "capturing", "non-capturing", 
                         "regex", "raw", "slashed-regex", "char-class"), "unknown args-class %s" % e
 
-                self.grammar._ARGS_OF_[_name] = _list
+                self.grammar._ARGS_OF_[_name] = (_list, postprocessor)
+
+            #elif tokens[0] == ':test':
+                #_name = tokens[1]
+                #self.grammar._TMP_['test'] = 
+
 
  #           elif tokens[0] == ":-sigspace":
  #               self.grammar._SIGSPACE_ = False
@@ -551,6 +596,15 @@ class GrammarCompiler:
 #            elif tokens[0] in (":+backtrack", ":+backtracking"):
 #                self.grammar._BACKTRACK_ = True
 #
+            elif tokens[0] in (":debug",):
+                self.grammar._TMP_['debug'] = 1
+
+            elif tokens[0] in (":+debug",):
+                self.grammar._TMP_['debugging'] = 1
+
+            elif tokens[0] in (":-debug",):
+                self.grammar._TMP_['debugging'] = 0
+
             elif tokens[0] in (":-skip",):
                 self.gramamr._SKIP_ = False
 
@@ -606,7 +660,7 @@ class GrammarCompiler:
 
 
 
-    def complex_rule(self, name, tokens):
+    def complex_rule(self, name, tokens, backtrack=False):
         '''handle complex rule
 
         Example::
@@ -615,7 +669,15 @@ class GrammarCompiler:
 
              <.foo{ first second third }>
         '''
-        args_of = self.grammar._ARGS_OF_.get(name, ['raw'])
+        args_of, postprocessor = self.grammar._ARGS_OF_.get(name, (['raw'], None))
+        if not args_of: args_of = ['raw']
+
+        if postprocessor is not None:
+            m, c = postprocessor.rsplit('.', 1)
+            x = {}
+            exec "from %s import %s" % (m,c) in x
+            postprocessor = x[c]
+
         rule = []
         while tokens:
             tok = tokens.pop(0)
@@ -623,12 +685,21 @@ class GrammarCompiler:
                 if len(tok) not in (2, 3):
                     raise GrammarError(self,
                        "unexpected token %(token)s", token=tok)
-                if tok[-1] in '?*+': quantifier = tok[-1] ; tok = tok[:-1]
+                quantifier, tok = self.quantifier(tok)
+               # if tok[-1] in '?*+': quantifier = tok[-1] ; tok = tok[:-1]
 
-                rule = NonCapturing(name, tuple(rule))
+                if postprocessor is None:
+                    args = tuple(rule)
+                else:
+                    args = postprocessor(self, name, rule, backtrack=backtrack)
+
+                if isinstance(args, Rule):
+                    rule = args
+                else:
+                    rule = NonCapturing(name, args, backtrack=backtrack)
 
                 if quantifier:
-                    return Quantified(rule, quantifier)
+                    return Quantified(rule, quantifier, backtrack=backtrack)
                 else:
                     return rule
 
@@ -637,11 +708,11 @@ class GrammarCompiler:
             for a in args_of:
                 if a == 'string' and self.rule_string(rule, tok):
                     break
-                elif a == 'capturing' and self.rule_capturing(rule, tok):
+                elif a == 'capturing' and self.rule_capturing(rule, tok, backtrack=backtrack):
                     break
                 elif (
                     a == 'non-capturing' \
-                    and self.rule_non_capturing(rule, tok)
+                    and self.rule_non_capturing(rule, tok, backtrack=backtrack)
                     ): break
                 elif a == 'slashed-regex' and tok[0] == "/" and tok[-1] == "/":
                     self.rule_regex(rule, tok[1:-1])
@@ -660,11 +731,12 @@ class GrammarCompiler:
 
         raise GrammarError(self, "unexpected end of token stream: expecting '}>'")
 
-    def quantified(self, rule, quantifier):
+    def quantified(self, rule, quantifier, backtrack=False):
         '''return a quantified rule'''
 
         if quantifier:
-            return Quantified(rule, quantifier)
+            return Quantified(rule, quantifier, backtrack=backtrack)
+
         return rule
 
     def quantifier(self, tok):
@@ -681,33 +753,33 @@ class GrammarCompiler:
 
         return quantifier, tok
 
-    def rule_non_capturing(self, rule, tok):
+    def rule_non_capturing(self, rule, tok, backtrack=False):
         if not self.RE_TRANSP_RULE.match(tok): return False
 
         quantifier, tok = self.quantifier(tok)
                 
         name = tok[2:-1]
-        r = NonCapturing(name)
-        rule.append(self.quantified(r, quantifier))
+        r = NonCapturing(name, backtrack=backtrack)
+        rule.append(self.quantified(r, quantifier, backtrack=backtrack))
 
         self._expect_rule(name)
 
         return True
 
-    def rule_capturing(self, rule, tok):
+    def rule_capturing(self, rule, tok, backtrack=False):
 
         if not self.RE_NAMED_RULE.match(tok): return False
 
         quantifier, tok = self.quantifier(tok)
         name = tok[1:-1]
 
-        rule.append(self.quantified(Capturing(name), quantifier))
+        rule.append(self.quantified(Capturing(name, backtrack=backtrack), quantifier, backtrack=backtrack))
 
         self._expect_rule(name)
 
         return True
 
-    def rule_with_arg(self, rule, tok):
+    def rule_with_arg(self, rule, tok, backtrack=False):
         if not self.RE_ARG_RULE.match(tok): return False
 
         quantifier, tok = self.quantifier(tok)
@@ -721,16 +793,16 @@ class GrammarCompiler:
             _args = re.compile(_args[1:-1])
 
         if _named:
-            r = Capturing(_name, _args)
+            r = Capturing(_name, _args, backtrack = backtrack)
         else:
-            r = NonCapturing(_name, _args)
+            r = NonCapturing(_name, _args, backtrack = backtrack)
 
-        rule.append(self.quantified(r, quantifier))
+        rule.append(self.quantified(r, quantifier, backtrack=backtrack))
 
         self._expect_rule(name)
         return True
 
-    def rule_single_quoted_string(self, rule, tok):
+    def rule_single_quoted_string(self, rule, tok, backtrack=None):
         if not self.RE_SQ_STRING.match(tok): return False
 
         s = re.sub(r"\\(\\|')", r'\1', tok[1:-1])
@@ -738,7 +810,7 @@ class GrammarCompiler:
 
         return True
        
-    def rule_double_quoted_string(self, rule, tok):
+    def rule_double_quoted_string(self, rule, tok, backtrack=None):
         if not self.RE_DQ_STRING.match(tok): return False
 
         s = re.sub(r'\\(\\|")', r'\1', tok[1:-1])
@@ -746,22 +818,27 @@ class GrammarCompiler:
 
         return True
 
-    def rule_string(self, rule, tok):
+    def rule_string(self, rule, tok, backtrack=False):
         if self.rule_single_quoted_string(rule, tok): return True
         if self.rule_double_quoted_string(rule, tok): return True
         return False
 
-    def any_rule(self, rule, tok, tokens):
+    def whitespace(self, tokens, backtrack=False):
+        '''create a whitespace rule'''
+        result = self.sequence(tokens, backtrack=backtrack, sigspace=False)
+        return result
+
+    def any_rule(self, rule, tok, tokens, backtrack=False):
         if self.RE_COMPLEX_RULE.match(tok):
-            rule.append(self.complex_rule(tok[1:-1], tokens))
+            rule.append(self.complex_rule(tok[1:-1], tokens, backtrack))
 
-        elif self.rule_non_capturing(rule, tok):
+        elif self.rule_non_capturing(rule, tok, backtrack):
             return
 
-        elif self.rule_capturing(rule, tok):
+        elif self.rule_capturing(rule, tok, backtrack):
             return
 
-        elif self.rule_with_arg(rule, tok):
+        elif self.rule_with_arg(rule, tok, backtrack=backtrack):
             return
 
         elif self.rule_string(rule, tok):
@@ -779,8 +856,12 @@ class GrammarCompiler:
             finally:
                 pass
    
-    def sequence(self, tokens):
-        sigspace = self.grammar._SIGSPACE_
+    def sequence(self, tokens, backtrack=None, sigspace=None):
+        if sigspace is None:
+            sigspace = self.grammar._SIGSPACE_
+        if backtrack is None:
+            sigspace = self.grammar._BACKTRACK_
+
         if isinstance(tokens, basestring): tokens = tokens.split()
         rule = []
 
@@ -789,7 +870,7 @@ class GrammarCompiler:
 
             quantifier = None
             if tok == '[':
-                rule.append(self.alternatives(tokens))
+                rule.append(self.alternatives(tokens, backtrack, sigspace))
                 continue
 
             elif tok == '|' or tok.startswith(']'):
@@ -798,7 +879,7 @@ class GrammarCompiler:
                     tokens.insert(0, tok) # push back token
                     break
 
-            self.any_rule(rule, tok, tokens)
+            self.any_rule(rule, tok, tokens, backtrack)
 
 
         if len(rule) == 1:
@@ -831,7 +912,7 @@ class GrammarCompiler:
 
             for r in rule:
                 new_rule.append(r)
-            
+
                 if is_sigspace(new_rule[-1]):
                     self.any_rule(new_rule, self.grammar._SIGSPACE_RULE_[0],
                         self.grammar._SIGSPACE_RULE_[1:])
@@ -839,21 +920,21 @@ class GrammarCompiler:
 
             rule = new_rule
 
-        return Sequence(rule, backtrack=self.grammar._BACKTRACK_)
+        return Sequence(rule, backtrack=backtrack)
 
-    def rule(self, tokens):
+    def rule(self, tokens, backtrack=False, sigspace=False):
         '''convenience call from outside'''
 
-        result = self.sequence(tokens)
+        result = self.sequence(tokens, backtrack=backtrack, sigspace=sigspace)
         assert not isinstance(result, Sequence)
         return result
 
-    def alternatives(self, tokens):
+    def alternatives(self, tokens, backtrack=False, sigspace=False):
         if isinstance(tokens, basestring): tokens = tokens.split()
         rule = []
         quantifier = None
         while tokens:
-            rule.append(self.sequence(tokens))
+            rule.append(self.sequence(tokens, backtrack=backtrack, sigspace=sigspace))
             if tokens:
                 tok = tokens.pop(0)
                 if tok == '|': continue
@@ -864,10 +945,10 @@ class GrammarCompiler:
         if len(rule) == 1:
            rule = rule[0]
         else:
-            rule = Alternative(rule, backtrack=self.grammar._BACKTRACK_)
+            rule = Alternative(rule, backtrack=backtrack)
 
         if quantifier:
-            rule = Quantified(rule, quantifier, backtrack=self.grammar._BACKTRACK_)
+            rule = Quantified(rule, quantifier, backtrack=backtrack)
         return rule
 
     def create_grammar(self, name, extends):
@@ -883,16 +964,25 @@ class GrammarCompiler:
         return self.grammar_factory(name, bases, {})
 
     def setup(self, G):
-        '''do some postprocessing'''
+        '''do some postprocessing
+
+        it is intended to classify, if a rule is terminal or if a rule is
+        recursive.
+
+        '''
+
 #        if 'ParseActions' not in G._PARSE_ACTIONS_:
 #            from .actions import ParseActions
 #            G._PARSE_ACTIONS_['ParseActions'] = ParseActions()
 
         for n in dir(G):
             if n.startswith('_') and n.endswith('_') and len(n) > 2: continue
+
             r = getattr(G, n)
             if isinstance(r, Rule):
-                r.setup(G)
+
+                r.setup(G, [ ])
+
                 if n in r.called:
                     r.recursive = n
 
@@ -937,7 +1027,8 @@ class GrammarCompiler:
 
             is_doc_string = self.is_doc_string
 
-        if isinstance(input, file):
+        if hasattr(input, 'read'):
+        #if isinstance(input, file):
             input = input.read()
 
         # apply filter to string 
@@ -988,9 +1079,9 @@ class GrammarCompiler:
         if self.grammar._EXPECTED_:
             raise GrammarError( self,
                 "following rules are expected, but not defined in grammar "
-                "%s:\n"%g.__name__ +
+                "%s:\n" % self.grammar.__name__ +
                 ''.join([ "  %s at line %s\n" %(k,v) 
-                for k,v in g._EXPECTED_.items()])
+                for k,v in self.grammar._EXPECTED_.items()])
             )
 
         self.setup(self.grammar)
