@@ -1,4 +1,4 @@
-#
+#l
 # Copyright 2012 by Kay-Uwe (Kiwi) Lorenz
 # Published under New BSD License, see LICENSE.txt for details.
 #
@@ -208,8 +208,15 @@ class String(Terminal):
         return MatchObject(start=p, end=p+self.slen, string=s, lexems=[])
 
     def match(self, P, s, start=0, end=None):
-        if s.startswith(self.data, start, end):
-            return MatchObject(start=start, end=start+self.slen, string=s)
+        # following code does not work with all python versions (in case end is None):
+        #   if s.startswith(self.data, start, end):
+        # So we have to differentiate:
+        if end:
+            if s.startswith(self.data, start, end):
+                return MatchObject(start=start, end=start+self.slen, string=s)
+        else:
+            if s.startswith(self.data, start):
+                return MatchObject(start=start, end=start+self.slen, string=s)
            
         return None
 
@@ -652,7 +659,7 @@ class Alternative(Rule):
         self.data = alternatives
 
     def __str__(self):
-        return " | ".join([str(x) for x in self.data])
+        return "[ " + " | ".join([str(x) for x in self.data]) + " ]"
 
     def rfind(self, P, s, start=0, end=None):
         assert self.is_terminal
@@ -701,7 +708,7 @@ class Alternative(Rule):
         self.called = called
                 
 
-CTXVAR_RE = re.compile(r'\$<[^>]*>')
+CTXVAR_RE = re.compile(r'\$<(?P<var>[^:>]*)(?::(?P<default>[^>]*))?>')
 class Regex(Terminal):
     '''match a regex.
 
@@ -733,10 +740,26 @@ class Regex(Terminal):
         if not isinstance(regex, basestring):
             return regex
         else:
-            regex = CTXVAR_RE.sub(
-                lambda m: re.escape(str(ctx(m.group(0)[2:-1]))), regex)
+            def replacement(m):
+                default = m.group('default')
+                if default is not None:
+                    s = ctxget( m.group('var'), default=default)
+                else:
+                    s = ctx(m.group('var'))
 
-            self.log_debug("%30.30s (dyn)", repr(regex))
+                self.log_debug("s1: %s\n", s)
+
+                if not isinstance(s, basestring):
+                    s = str(s)
+                    self.log_debug("s2: %s\n", s)
+
+                return re.escape( s )
+
+            self.log_debug("%30.30s (dyn old)", repr(regex))
+
+            regex = CTXVAR_RE.sub(replacement, regex)
+
+            self.log_debug("%30.30s (dyn new)", repr(regex))
 
             return re.compile(regex)
 
